;;;;
;;;; Hacky modification/additions to CL-WHO to print forms to a stream, or 
;;;; a string. Tested against CL-WHO 0.9.1.
;;;; It's probably best if you put these into a different package...!
;;;;
;;;; Modifications/additions by Richard Newman, December 2005. Consider this
;;;; code contributed to CL-WHO.
;;;;

(in-package :cl-who)

(defun tree-to-commands-writer-aux (tree stream)
  (declare (optimize speed space))
  "Print the HTML in TREE to STREAM. Utility function used by
   TREE-TO-STREAM. Hacked out of CL-WHO core."
  (let ((in-string t)
        collector
        string-collector)
    (flet ((emit-string-collector ()
             "Generate a WRITE-STRING statement for what is currently
              in STRING-COLLECTOR."
             (write-string
                   (string-list-to-string (nreverse string-collector))
                   stream))
           (tree-to-commands-aux-internal (tree)
             "Same as TREE-TO-COMMANDS-AUX but with closed-over STREAM
              for REPLACE-HTM."
             (tree-to-commands-writer-aux tree stream)))
      (unless (listp tree)
        (return-from tree-to-commands-writer-aux tree))
      (loop for element in tree
            do (cond ((and in-string (stringp element))
                       ;; this element is a string and the last one
                       ;; also was (or this is the first element) -
                       ;; collect into STRING-COLLECTOR
                       (push element string-collector))
                     ((stringp element)
                       ;; the last one wasn't a string so we start
                       ;; with an empty STRING-COLLECTOR
                       (setq string-collector (list element)
                             in-string t))
                     (string-collector
                       ;; not a string but STRING-COLLECTOR isn't
                       ;; empty so we have to emit the collected
                       ;; strings first
                       (push (emit-string-collector) collector)
                       (setq in-string nil
                             string-collector '())
                       ;; collect this element but walk down the
                       ;; subtree first
                       (push (replace-htm element #'tree-to-commands-aux-internal)
                             collector))
                     (t
                       ;; not a string and empty STRING-COLLECTOR
                       (push (replace-htm element #'tree-to-commands-aux-internal)
                             collector)))
            finally (return (if string-collector
                              ;; finally empty STRING-COLLECTOR if
                              ;; there's something in it
                              (nreverse (cons (emit-string-collector)
                                              collector))
                              (nreverse collector)))))))

(defun tree-to-stream (tree &optional (stream *standard-output*) prologue)
  (declare (optimize speed space))
  "Transforms an HTML tree into HTML on STREAM.
   Provide t for prologue to use the default. A string will be printed
   before the output. nil means no prologue."
  ;; use TREE-TO-TEMPLATE, then TREE-TO-COMMANDS-AUX, and finally
  ;; replace the special symbols ESC, STR, FMT, and HTM
  (let ((*print-pretty* nil))  ; this *may* provide a speedup if FORMAT is called within.
    (declare (type cons tree)
             (type stream stream))
    (when prologue
      (if (stringp prologue)
        (princ prologue stream)
        (princ *prologue* stream)))
    (apply-to-tree #'(lambda (x)
                       (case (first x)
                         ((esc)
                          (write-string
                            (escape-string (second x))
                            stream))
                         ((str)
                          (princ (second x) stream))
                         ((fmt)
                          (format stream (rest x)))))
                   #'(lambda (x)
                       (and (consp x)
                            (member (first x)
                                    '(esc str fmt)
                                    :test #'eq)))
                   (apply-to-tree (constantly 'progn)
                                  #'(lambda (x)
                                      (and (atom x)
                                           (eq x 'htm)))
                                  (tree-to-commands-writer-aux
                                    (cons 'htm
                                          ;; For some reason we need to wrap this in (htm).
                                          (tree-to-template `(htm ,tree)))
                                    stream)))
    t))    ;; return T rather than the list we've built...

(defun tree-to-string (tree)
  (with-output-to-string (x)
    (tree-to-stream tree x)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (export '(tree-to-string tree-to-stream) (find-package :cl-who)))
